==> go-test-framework-223.8836.34-sources.jar: com/goide/GoCodeInsightFixtureTestCase.java <==
/*
 * Copyright 2013-2016 Sergey Ignatov, Alexander Zolotov, Florin Patan and 2016-2021 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.goide;

import com.goide.documentation.GoDocumentationProvider;
import com.goide.formatter.settings.GoCodeStyleSettings;
import com.goide.highlighting.GoAnnotator;
import com.goide.project.GoModuleSettings;
import com.goide.psi.*;
import com.goide.vendor.GoVendorDirectoriesCollector;
import com.intellij.application.options.CodeStyle;
import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.codeInsight.lookup.LookupManager;
import com.intellij.codeInsight.template.impl.TemplateManagerImpl;
import com.intellij.codeInsight.template.impl.TemplateState;
import com.intellij.injected.editor.VirtualFileWindow;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.application.WriteAction;
import com.intellij.openapi.application.impl.NonBlockingReadActionImpl;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.SelectionModel;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Comparing;
import com.intellij.openapi.util.Condition;
import com.intellij.openapi.util.Conditions;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VfsUtilCore;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileFilter;
import com.intellij.psi.*;
import com.intellij.psi.codeStyle.CommonCodeStyleSettings;
import com.intellij.psi.impl.DebugUtil;
import com.intellij.psi.impl.PsiManagerImpl;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.refactoring.util.CommonRefactoringUtil;
import com.intellij.testFramework.PlatformTestUtil;
import com.intellij.testFramework.TestDataFile;
import com.intellij.testFramework.UsefulTestCase;
import com.intellij.testFramework.fixtures.BasePlatformTestCase;
import com.intellij.testFramework.fixtures.CodeInsightTestFixture;
import com.intellij.testFramework.fixtures.impl.CodeInsightTestFixtureImpl;
import com.intellij.util.ObjectUtils;
import com.intellij.util.PathUtil;
import com.intellij.util.ThreeState;
import com.intellij.util.ThrowableRunnable;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.indexing.FileContentImpl;
import com.intellij.util.indexing.IndexingDataKeys;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;

abstract public class GoCodeInsightFixtureTestCase extends BasePlatformTestCase {

  protected PsiFile configure(@NotNull String relativePath, @NotNull String text) {
    PsiFile file = myFixture.addFileToProject(relativePath, text);
    myFixture.configureFromExistingVirtualFile(file.getVirtualFile());
    return file;
  }

  protected void checkResult(@TestDataFile @NotNull String filePath, @NotNull String text) {
    myFixture.checkResult(filePath, text, false);
  }

  @NotNull
  protected PsiFile testHighlighting(@NotNull String relativePath, @NotNull String text) {
    PsiFile file = configure(relativePath, text);
    failOnFileLoading();
    myFixture.checkHighlighting();
    return file;
  }

  @NotNull
  protected PsiFile testHighlighting(@TestDataFile @NotNull String filePath) {
    PsiFile file = myFixture.configureByFile(filePath);
    failOnFileLoading();
    myFixture.checkHighlighting();
    return file;
  }

  protected void testHighlighting(boolean checkWarnings,
                                  boolean checkInfos,
                                  boolean checkWeakWarnings,
                                  @TestDataFile @NotNull String filePath) {
    myFixture.configureByFile(filePath);
    failOnFileLoading();
    myFixture.checkHighlighting(checkWarnings, checkInfos, checkWeakWarnings);
  }

  @NotNull
  protected PsiFile doTestHighlighting() {
    return testHighlighting(getTestName(true) + ".go");
  }

  protected static String buildStubTreeText(@NotNull Project project,
                                            @NotNull VirtualFile file,
                                            @NotNull String fileContent,
                                            boolean checkErrors) {
    String path = file.getPath();
    PsiFile psi = PsiFileFactory.getInstance(project).createFileFromText(file.getName(), file.getFileType(), fileContent);
    if (checkErrors) {
      assertFalse(path + " contains error elements", DebugUtil.psiToString(psi, false).contains("PsiErrorElement"));
    }
    String full = DebugUtil.stubTreeToString(GoFileElementType.INSTANCE.getBuilder().buildStubTree(psi));
    psi.putUserData(IndexingDataKeys.VIRTUAL_FILE, file);
    FileContentImpl content = (FileContentImpl)FileContentImpl.createByContent(file, fileContent.getBytes(file.getCharset()));
    PsiFile psiFile = content.getPsiFile();
    String fast = DebugUtil.stubTreeToString(GoFileElementType.INSTANCE.getBuilder().buildStubTree(psiFile));
    if (!Comparing.strEqual(full, fast)) {
      System.err.println(path);
      UsefulTestCase.assertSameLines(full, fast);
    }
    return fast;
  }

  @NotNull
  protected PsiElement findElementAtCaretOrInSelection() {
    SelectionModel selectionModel = myFixture.getEditor().getSelectionModel();
    if (selectionModel.hasSelection()) {
      PsiElement left = myFixture.getFile().findElementAt(selectionModel.getSelectionStart());
      PsiElement right = myFixture.getFile().findElementAt(selectionModel.getSelectionEnd() - 1);
      assertNotNull(left);
      assertNotNull(right);
      return Objects.requireNonNull(PsiTreeUtil.findCommonParent(left, right));
    }
    else {
      return Objects.requireNonNull(myFixture.getFile().findElementAt(myFixture.getEditor().getCaretModel().getOffset()));
    }
  }

  @NotNull
  protected static List<String> getExpectedData(@NotNull PsiElement element) {
    PsiElement parent = PsiTreeUtil.getNonStrictParentOfType(element, GoStatement.class, GoSignature.class, GoMethodSpec.class,
                                                             GoFieldDeclaration.class, GoTopLevelDeclaration.class);
    // var and const declarations are GoTopLevelDeclaration so we need to get parent statement in this case
    parent = parent instanceof GoTopLevelDeclaration && parent.getParent() instanceof GoStatement ? parent.getParent() : parent;
    List<PsiComment> comments = parent != null ? GoDocumentationProvider.getAfterComments(parent) : Collections.emptyList();
    String comment = !comments.isEmpty() ? comments.get(0).getText() : null;
    String text = comment != null ? StringUtil.trimStart(StringUtil.trimEnd(comment, "*/"), "/*") : null;
    if (text == null) {
      fail("test data should contain comment after a parent statement, signature, method spec, field declaration, or top-level declaration " +
           "in the format: '/* <value1>, <value2>, ... */'");
    }
    return text.isEmpty() ? Collections.emptyList() : ContainerUtil.map(text.split(","), String::trim);
  }

  protected @Nullable PsiElement findElementByText(@NotNull String text) {
    int pos = myFixture.getDocument(myFixture.getFile()).getText().indexOf(text);
    return pos >= 0 ? myFixture.getFile().findElementAt(pos) : null;
  }

  @Override
  protected void setUp() throws Exception {
    super.setUp();
    if (runInDispatchThread()) setupGoSupport();
    else ApplicationManager.getApplication().invokeAndWait(this::setupGoSupport, ModalityState.NON_MODAL);
  }

  protected void setupGoSupport() {
    GoTestUtil.setupGoSupport(this, getProject(), getModule(), true, null);
  }

  @Override
  protected void tearDown() throws Exception {
    try {
      UIUtil.invokeAndWaitIfNeeded((ThrowableRunnable<?>)() -> LookupManager.getInstance(myFixture.getProject()).hideActiveLookup());
      Editor editor = myFixture.getEditor();
      TemplateState state = editor != null ? TemplateManagerImpl.getTemplateState(editor) : null;
      if (state != null) {
        Disposer.dispose(state);
      }
      ((PsiManagerImpl)myFixture.getPsiManager()).setAssertOnFileLoadingFilter(VirtualFileFilter.NONE, getTestRootDisposable());
    }
    catch (Throwable e) {
      addSuppressedException(e);
    }
    finally {
      super.tearDown();
    }
  }

  @Override
  protected String getTestDataPath() {
    return GoTestUtil.getTestDataPath(getBasePath()).getAbsolutePath();
  }

  @NotNull
  protected static String loadText(@NotNull VirtualFile file) {
    try {
      return StringUtil.convertLineSeparators(VfsUtilCore.loadText(file));
    }
    catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  protected void disableVendoring() {
    GoModuleSettings.getInstance(myFixture.getModule()).setVendoringEnabled(ThreeState.NO);
  }

  protected static String normalizeCode(@NotNull String codeBefore) {
    StringBuilder result = new StringBuilder("package main\nfunc main() {\n");
    if ("\n".equals(codeBefore)) {
      result.append(codeBefore);
    }
    for (String line : StringUtil.splitByLines(codeBefore, false)) {
      result.append('\t').append(line).append('\n');
    }
    result.append("}");
    return result.toString();
  }

  protected void failOnFileLoading(VirtualFile... filesToIgnore) {
    failOnFileLoading(this, myFixture, filesToIgnore);
  }

  public static void failOnFileLoading(@NotNull UsefulTestCase testCase, @NotNull CodeInsightTestFixture fixture, VirtualFile... ignore) {
    if (GoTestUtil.findAnnotation(testCase, StubDestroyer.class) != null) {
      return;
    }
    HashSet<VirtualFile> filesToIgnore = ContainerUtil.newHashSet(ignore);
    VirtualFileFilter filter = file -> !(file instanceof VirtualFileWindow || filesToIgnore.contains(file));
    ((PsiManagerImpl)fixture.getPsiManager()).setAssertOnFileLoadingFilter(filter, fixture.getTestRootDisposable());
    if (fixture instanceof CodeInsightTestFixtureImpl) {
      ((CodeInsightTestFixtureImpl)fixture).setVirtualFileFilter(filter);
    }
  }

  protected void applySingleQuickFix(@NotNull String quickFixNameHint) {
    applySingleQuickFix(quickFixNameHint, null);
  }

  protected void applySingleQuickFix(@NotNull String quickFixName, @Nullable Condition<? super IntentionAction> quickFixSelector) {
    applySingleQuickFix(quickFixName, false, quickFixSelector);
  }

  protected void applySingleQuickFix(@NotNull String quickFixName,
                                     boolean checkPreview,
                                     @Nullable Condition<? super IntentionAction> quickFixSelector) {
    List<IntentionAction> availableIntentions = myFixture.filterAvailableIntentions(quickFixName);
    IntentionAction action = ContainerUtil.find(availableIntentions, ObjectUtils.notNull(quickFixSelector, Conditions.alwaysTrue()));
    if (action == null) {
      String allIntentions = StringUtil.join(myFixture.getAvailableIntentions(), IntentionAction::getText, ", ");
      fail("Cannot find `" + quickFixName + "` fix. Available intentions: [" + allIntentions + "]");
    }
    if (checkPreview) {
      myFixture.checkPreviewAndLaunchAction(action);
    }
    else {
      myFixture.launchAction(action);
    }
  }

  protected void doNotConsiderVendorAsLibrary() {
    GoVendorDirectoriesCollector.getInstance(myFixture.getProject()).disableInTest(getTestRootDisposable());
  }

  protected void ensureVendorLibrariesAttached() {
    GoVendorDirectoriesCollector.getInstance(myFixture.getProject()).recalculateVendorDirectories();
    PlatformTestUtil.dispatchAllInvocationEventsInIdeEventQueue();
  }

  public void assertLoadingAst(@NotNull Runnable runnable, @NotNull String filePath) {
    assertThrows(Throwable.class, () -> {
        try {
          runnable.run();
        }
        catch (AssertionError e) {
          e = unwrapCause(e);
          String message = e.getMessage();
          assertTrue("Wrong assertion: " + message, message.contains("Access to tree elements not allowed"));
          assertTrue(filePath + " wasn't loaded: " + message, message.contains(PathUtil.toSystemDependentName(filePath)));
          throw e;
        }
      }
    );
  }

  @NotNull
  private static AssertionError unwrapCause(AssertionError e) {
    Throwable cause = e.getCause();
    while (e != cause && cause instanceof AssertionError) {
      e = (AssertionError)cause;
      cause = e.getCause();
    }
    return e;
  }

  public void enableErrorChecks(GoAnnotator.InspectionVisitorConstructor @NotNull ... constructors) {
    GoAnnotator.enableChecks(getTestRootDisposable(), constructors);
  }

  protected static void assertRefactoringError(@NotNull Runnable test) {
    assertRefactoringError("", test);
  }

  protected static void assertRefactoringError(@NotNull String errorMessagePart, @NotNull Runnable test) {
    try {
      test.run();
      fail("Test is expected to fail with a refactoring error");
    } catch (CommonRefactoringUtil.RefactoringErrorHintException e) {
      assertTrue(e.getMessage().contains(errorMessagePart));
    }
  }

  // This interface is redundant, but simplifies use of runWithCustomCodeStyle:
  // IDE inserts proper parameter names instead of settings1, settings2.
  public interface SettingsModifier extends BiConsumer<CommonCodeStyleSettings, GoCodeStyleSettings> {
    @Override
    void accept(CommonCodeStyleSettings commonSettings, GoCodeStyleSettings goSettings);
  }

  public void runWithCustomCodeStyle(@NotNull SettingsModifier settingsModifier) {
    CodeStyle.doWithTemporarySettings(myFixture.getProject(), CodeStyle.getDefaultSettings(), settings -> {
      CommonCodeStyleSettings commonSettings = settings.getCommonSettings(GoLanguage.INSTANCE);
      GoCodeStyleSettings goSettings = settings.getCustomSettings(GoCodeStyleSettings.class);
      settingsModifier.accept(commonSettings, goSettings);
    });
  }

  protected PsiFile configureScratchFile(@NotNull String text) throws IOException {
    int offset = StringUtil.indexOf(text, "<caret>");
    PsiFile file = configure("file.go", "<selection>" + text + "</selection>");

    GoTestUtil.newScratchFile(myFixture.getEditor());
    VirtualFile scratch = GoTestUtil.getScratchVirtualFile(getProject(), getTestRootDisposable());
    assertNotNull(scratch);
    WriteAction.runAndWait(() -> VfsUtil.saveText(scratch, text));
    WriteAction.runAndWait(() -> file.delete());

    myFixture.configureFromExistingVirtualFile(scratch);
    myFixture.getEditor().getCaretModel().moveToOffset(offset);
    NonBlockingReadActionImpl.waitForAsyncTaskCompletion();
    return PsiManager.getInstance(getProject()).findFile(scratch);
  }
}

==> go-test-framework-231.9161.41-sources.jar: com/goide/GoCodeInsightFixtureTestCase.java <==
/*
 * Copyright 2013-2016 Sergey Ignatov, Alexander Zolotov, Florin Patan and 2016-2021 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.goide;

import com.goide.documentation.GoDocumentationProvider;
import com.goide.formatter.settings.GoCodeStyleSettings;
import com.goide.highlighting.GoAnnotator;
import com.goide.project.GoModuleSettings;
import com.goide.psi.*;
import com.goide.vendor.GoVendorDirectoriesCollector;
import com.intellij.application.options.CodeStyle;
import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.codeInsight.lookup.LookupManager;
import com.intellij.codeInsight.template.impl.TemplateManagerImpl;
import com.intellij.codeInsight.template.impl.TemplateState;
import com.intellij.injected.editor.VirtualFileWindow;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.application.WriteAction;
import com.intellij.openapi.application.impl.NonBlockingReadActionImpl;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.SelectionModel;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Comparing;
import com.intellij.openapi.util.Condition;
import com.intellij.openapi.util.Conditions;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VfsUtilCore;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileFilter;
import com.intellij.psi.*;
import com.intellij.psi.codeStyle.CommonCodeStyleSettings;
import com.intellij.psi.impl.DebugUtil;
import com.intellij.psi.impl.PsiManagerImpl;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.refactoring.util.CommonRefactoringUtil;
import com.intellij.testFramework.PlatformTestUtil;
import com.intellij.testFramework.TestDataFile;
import com.intellij.testFramework.UsefulTestCase;
import com.intellij.testFramework.fixtures.BasePlatformTestCase;
import com.intellij.testFramework.fixtures.CodeInsightTestFixture;
import com.intellij.testFramework.fixtures.impl.CodeInsightTestFixtureImpl;
import com.intellij.util.ObjectUtils;
import com.intellij.util.PathUtil;
import com.intellij.util.ThreeState;
import com.intellij.util.ThrowableRunnable;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.indexing.FileContentImpl;
import com.intellij.util.indexing.IndexingDataKeys;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;

abstract public class GoCodeInsightFixtureTestCase extends BasePlatformTestCase {

  protected PsiFile configure(@NotNull String relativePath, @NotNull String text) {
    PsiFile file = myFixture.addFileToProject(relativePath, text);
    myFixture.configureFromExistingVirtualFile(file.getVirtualFile());
    return file;
  }

  protected void checkResult(@TestDataFile @NotNull String filePath, @NotNull String text) {
    myFixture.checkResult(filePath, text, false);
  }

  @NotNull
  protected PsiFile testHighlighting(@NotNull String relativePath, @NotNull String text) {
    PsiFile file = configure(relativePath, text);
    failOnFileLoading();
    myFixture.checkHighlighting();
    return file;
  }

  @NotNull
  protected PsiFile testHighlighting(@TestDataFile @NotNull String filePath) {
    PsiFile file = myFixture.configureByFile(filePath);
    failOnFileLoading();
    myFixture.checkHighlighting();
    return file;
  }

  protected void testHighlighting(boolean checkWarnings,
                                  boolean checkInfos,
                                  boolean checkWeakWarnings,
                                  @TestDataFile @NotNull String filePath) {
    myFixture.configureByFile(filePath);
    failOnFileLoading();
    myFixture.checkHighlighting(checkWarnings, checkInfos, checkWeakWarnings);
  }

  @NotNull
  protected PsiFile doTestHighlighting() {
    return testHighlighting(getTestName(true) + ".go");
  }

  protected static String buildStubTreeText(@NotNull Project project,
                                            @NotNull VirtualFile file,
                                            @NotNull String fileContent,
                                            boolean checkErrors) {
    String path = file.getPath();
    PsiFile psi = PsiFileFactory.getInstance(project).createFileFromText(file.getName(), file.getFileType(), fileContent);
    if (checkErrors) {
      assertFalse(path + " contains error elements", DebugUtil.psiToString(psi, false).contains("PsiErrorElement"));
    }
    String full = DebugUtil.stubTreeToString(GoFileElementType.INSTANCE.getBuilder().buildStubTree(psi));
    psi.putUserData(IndexingDataKeys.VIRTUAL_FILE, file);
    FileContentImpl content = (FileContentImpl)FileContentImpl.createByContent(file,
                                                                               () -> fileContent.getBytes(file.getCharset()),
                                                                               psi.getProject());
    PsiFile psiFile = content.getPsiFile();
    String fast = DebugUtil.stubTreeToString(GoFileElementType.INSTANCE.getBuilder().buildStubTree(psiFile));
    if (!Comparing.strEqual(full, fast)) {
      System.err.println(path);
      UsefulTestCase.assertSameLines(full, fast);
    }
    return fast;
  }

  @NotNull
  protected PsiElement findElementAtCaretOrInSelection() {
    SelectionModel selectionModel = myFixture.getEditor().getSelectionModel();
    if (selectionModel.hasSelection()) {
      PsiElement left = myFixture.getFile().findElementAt(selectionModel.getSelectionStart());
      PsiElement right = myFixture.getFile().findElementAt(selectionModel.getSelectionEnd() - 1);
      assertNotNull(left);
      assertNotNull(right);
      return Objects.requireNonNull(PsiTreeUtil.findCommonParent(left, right));
    }
    else {
      return Objects.requireNonNull(myFixture.getFile().findElementAt(myFixture.getEditor().getCaretModel().getOffset()));
    }
  }

  @NotNull
  protected static List<String> getExpectedData(@NotNull PsiElement element) {
    PsiElement parent = PsiTreeUtil.getNonStrictParentOfType(element, GoStatement.class, GoSignature.class, GoMethodSpec.class,
                                                             GoFieldDeclaration.class, GoTopLevelDeclaration.class);
    // var and const declarations are GoTopLevelDeclaration so we need to get parent statement in this case
    parent = parent instanceof GoTopLevelDeclaration && parent.getParent() instanceof GoStatement ? parent.getParent() : parent;
    List<PsiComment> comments = parent != null ? GoDocumentationProvider.getAfterComments(parent) : Collections.emptyList();
    String comment = !comments.isEmpty() ? comments.get(0).getText() : null;
    String text = comment != null ? StringUtil.trimStart(StringUtil.trimEnd(comment, "*/"), "/*") : null;
    if (text == null) {
      fail("test data should contain comment after a parent statement, signature, method spec, field declaration, or top-level declaration " +
           "in the format: '/* <value1>, <value2>, ... */'");
    }
    return text.isEmpty() ? Collections.emptyList() : ContainerUtil.map(text.split(","), String::trim);
  }

  protected @Nullable PsiElement findElementByText(@NotNull String text) {
    int pos = myFixture.getDocument(myFixture.getFile()).getText().indexOf(text);
    return pos >= 0 ? myFixture.getFile().findElementAt(pos) : null;
  }

  @Override
  protected void setUp() throws Exception {
    super.setUp();
    if (runInDispatchThread()) setupGoSupport();
    else ApplicationManager.getApplication().invokeAndWait(this::setupGoSupport, ModalityState.NON_MODAL);
  }

  protected void setupGoSupport() {
    GoTestUtil.setupGoSupport(this, getProject(), getModule(), true, null);
  }

  @Override
  protected void tearDown() throws Exception {
    try {
      UIUtil.invokeAndWaitIfNeeded((ThrowableRunnable<?>)() -> LookupManager.getInstance(myFixture.getProject()).hideActiveLookup());
      Editor editor = myFixture.getEditor();
      TemplateState state = editor != null ? TemplateManagerImpl.getTemplateState(editor) : null;
      if (state != null) {
        Disposer.dispose(state);
      }
      ((PsiManagerImpl)myFixture.getPsiManager()).setAssertOnFileLoadingFilter(VirtualFileFilter.NONE, getTestRootDisposable());
    }
    catch (Throwable e) {
      addSuppressedException(e);
    }
    finally {
      super.tearDown();
    }
  }

  @Override
  protected String getTestDataPath() {
    return GoTestUtil.getTestDataPath(getBasePath()).getAbsolutePath();
  }

  @NotNull
  protected static String loadText(@NotNull VirtualFile file) {
    try {
      return StringUtil.convertLineSeparators(VfsUtilCore.loadText(file));
    }
    catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  protected void disableVendoring() {
    GoModuleSettings.getInstance(myFixture.getModule()).setVendoringEnabled(ThreeState.NO);
  }

  protected static String normalizeCode(@NotNull String codeBefore) {
    StringBuilder result = new StringBuilder("package main\nfunc main() {\n");
    if ("\n".equals(codeBefore)) {
      result.append(codeBefore);
    }
    for (String line : StringUtil.splitByLines(codeBefore, false)) {
      result.append('\t').append(line).append('\n');
    }
    result.append("}");
    return result.toString();
  }

  protected void failOnFileLoading(VirtualFile... filesToIgnore) {
    failOnFileLoading(this, myFixture, filesToIgnore);
  }

  public static void failOnFileLoading(@NotNull UsefulTestCase testCase, @NotNull CodeInsightTestFixture fixture, VirtualFile... ignore) {
    if (GoTestUtil.findAnnotation(testCase, StubDestroyer.class) != null) {
      return;
    }
    HashSet<VirtualFile> filesToIgnore = ContainerUtil.newHashSet(ignore);
    VirtualFileFilter filter = file -> !(file instanceof VirtualFileWindow || filesToIgnore.contains(file));
    ((PsiManagerImpl)fixture.getPsiManager()).setAssertOnFileLoadingFilter(filter, fixture.getTestRootDisposable());
    if (fixture instanceof CodeInsightTestFixtureImpl) {
      ((CodeInsightTestFixtureImpl)fixture).setVirtualFileFilter(filter);
    }
  }

  protected void applySingleQuickFix(@NotNull String quickFixNameHint) {
    applySingleQuickFix(quickFixNameHint, null);
  }

  protected void applySingleQuickFix(@NotNull String quickFixNameHint, boolean checkPreview) {
    applySingleQuickFix(quickFixNameHint, checkPreview, null);
  }

  protected void applySingleQuickFix(@NotNull String quickFixName, @Nullable Condition<? super IntentionAction> quickFixSelector) {
    applySingleQuickFix(quickFixName, true, quickFixSelector);
  }

  protected void applySingleQuickFix(@NotNull String quickFixName,
                                     boolean checkPreview,
                                     @Nullable Condition<? super IntentionAction> quickFixSelector) {
    List<IntentionAction> availableIntentions = myFixture.filterAvailableIntentions(quickFixName);
    IntentionAction action = ContainerUtil.find(availableIntentions, ObjectUtils.notNull(quickFixSelector, Conditions.alwaysTrue()));
    if (action == null) {
      String allIntentions = StringUtil.join(myFixture.getAvailableIntentions(), IntentionAction::getText, ", ");
      fail("Cannot find `" + quickFixName + "` fix. Available intentions: [" + allIntentions + "]");
    }
    if (checkPreview) {
      myFixture.checkPreviewAndLaunchAction(action);
    }
    else {
      myFixture.launchAction(action);
    }
  }

  protected void doNotConsiderVendorAsLibrary() {
    GoVendorDirectoriesCollector.getInstance(myFixture.getProject()).disableInTest(getTestRootDisposable());
  }

  protected void ensureVendorLibrariesAttached() {
    GoVendorDirectoriesCollector.getInstance(myFixture.getProject()).recalculateVendorDirectories();
    PlatformTestUtil.dispatchAllInvocationEventsInIdeEventQueue();
  }

  public void assertLoadingAst(@NotNull Runnable runnable, @NotNull String filePath) {
    assertThrows(Throwable.class, () -> {
        try {
          runnable.run();
        }
        catch (AssertionError e) {
          e = unwrapCause(e);
          String message = e.getMessage();
          assertTrue("Wrong assertion: " + message, message.contains("Access to tree elements not allowed"));
          assertTrue(filePath + " wasn't loaded: " + message, message.contains(PathUtil.toSystemDependentName(filePath)));
          throw e;
        }
      }
    );
  }

  @NotNull
  private static AssertionError unwrapCause(AssertionError e) {
    Throwable cause = e.getCause();
    while (e != cause && cause instanceof AssertionError) {
      e = (AssertionError)cause;
      cause = e.getCause();
    }
    return e;
  }

  public void enableErrorChecks(GoAnnotator.InspectionVisitorConstructor @NotNull ... constructors) {
    GoAnnotator.enableChecks(getTestRootDisposable(), constructors);
  }

  protected static void assertRefactoringError(@NotNull Runnable test) {
    assertRefactoringError("", test);
  }

  protected static void assertRefactoringError(@NotNull String errorMessagePart, @NotNull Runnable test) {
    try {
      test.run();
      fail("Test is expected to fail with a refactoring error");
    } catch (CommonRefactoringUtil.RefactoringErrorHintException e) {
      assertTrue(e.getMessage().contains(errorMessagePart));
    }
  }

  // This interface is redundant, but simplifies use of runWithCustomCodeStyle:
  // IDE inserts proper parameter names instead of settings1, settings2.
  public interface SettingsModifier extends BiConsumer<CommonCodeStyleSettings, GoCodeStyleSettings> {
    @Override
    void accept(CommonCodeStyleSettings commonSettings, GoCodeStyleSettings goSettings);
  }

  public void runWithCustomCodeStyle(@NotNull SettingsModifier settingsModifier) {
    CodeStyle.doWithTemporarySettings(myFixture.getProject(), CodeStyle.getDefaultSettings(), settings -> {
      CommonCodeStyleSettings commonSettings = settings.getCommonSettings(GoLanguage.INSTANCE);
      GoCodeStyleSettings goSettings = settings.getCustomSettings(GoCodeStyleSettings.class);
      settingsModifier.accept(commonSettings, goSettings);
    });
  }

  protected PsiFile configureScratchFile(@NotNull String text) throws IOException {
    int offset = StringUtil.indexOf(text, "<caret>");
    PsiFile file = configure("file.go", "<selection>" + text + "</selection>");

    GoTestUtil.newScratchFile(myFixture.getEditor());
    VirtualFile scratch = GoTestUtil.getScratchVirtualFile(getProject(), getTestRootDisposable());
    assertNotNull(scratch);
    WriteAction.runAndWait(() -> VfsUtil.saveText(scratch, text));
    WriteAction.runAndWait(() -> file.delete());

    myFixture.configureFromExistingVirtualFile(scratch);
    myFixture.getEditor().getCaretModel().moveToOffset(offset);
    NonBlockingReadActionImpl.waitForAsyncTaskCompletion();
    return PsiManager.getInstance(getProject()).findFile(scratch);
  }
}
